from functools import wraps

overload_handlers_map = {}
__all__ = ['overload']


def extract_signature_function(fn):
    """
    Funzione per l'estrazione della firma della funzione passata come parametro

    :param fn:
    :return:
    """

    if not fn.__annotations__:
        raise TypeError(f'Missing annotation for {fn.__qualname__}')

    return fn.__annotations__


def is_matching_signature(args_tuple, signature, handler) -> bool:
    """
    Funzione per il controllo del match tra gli args/kwargs e la funzione passati

    :param args_tuple: Tuple that contains *args and **kwargs
    :param signature: Signature of decorated function or class method
    :param handler: Execution of decorated function or class method
    :return: Return True if find signature in handlers_map else False
    """
    args, kwargs = args_tuple

    if len(args) + len(kwargs) != len(handler.__code__.co_varnames):
        return False

    if len(args):
        mod = handler.__qualname__.replace(f".{handler.__name__}", "")
        for value, (_key, _type) in zip(args, signature.items()):
            is_method = mod in value.__class__.__qualname__
            if not is_method:
                # Here is a function
                if not isinstance(value, _type):
                    return False
            else:
                # Here is a class method
                if not isinstance(args[1], _type):
                    return False

    if len(kwargs):
        for arg_name in kwargs:
            if arg_name not in signature:
                return False

            if not isinstance(kwargs[arg_name], signature[arg_name]):
                return False

    return True


def signature_already_exists(searching_signature, handlers_map):
    """
    Checking the existence of the function signature passed with those saved in the "handlers_map" map

    :param searching_signature:
    :param handlers_map:
    :return:
    """
    for (signature, handler) in handlers_map:
        if searching_signature == signature:
            return True

    return False


def overload(fn):
    """
    MODULE NAME
        overload

    MODULE REFERENCE
        https://decoratorutilities.readthedocs.io/en/latest/rst_templates/Decorators/overloading_decorator.html

    Decorate your own function or class method with **@overload** decorator to define multiple functions with same name but with different parameters

    :param fn: Decorated function or class method to overload
    :return: Return the execution of function or class method
    """
    fn_name = f"{fn.__module__}.{fn.__qualname__}"

    if fn_name not in overload_handlers_map:
        overload_handlers_map[fn_name] = []

    current_signature = extract_signature_function(fn)
    if signature_already_exists(current_signature, overload_handlers_map[fn_name]):
        raise NameError(f'Signature already exists for {fn.__qualname__}')

    overload_handlers_map[fn_name].append((current_signature, fn))

    @wraps(fn)
    def wrapper(*args, **kwargs):
        for (signature, handler) in overload_handlers_map[fn_name]:
            if is_matching_signature((args, kwargs), signature, handler):
                return handler(*args, **kwargs)
        else:
            raise ValueError(f'No matching signature for {fn.__qualname__} with arguments: {args} {kwargs}')

    return wrapper



import datetime
from functools import wraps

__all__ = ['timeit']


def timeit(fn):
    """
    MODULE NAME
        timeit

    MODULE REFERENCE
        https://decoratorutilities.readthedocs.io/en/latest/rst_templates/Decorators/timeit_decorator.html

    Decorate your own function with **@timeit** decorator to monitoring execution time

    :param fn: Decorated function or class method to monitoring execution time
    :return:
    """

    @wraps(fn)
    def wrapper(*args, **kwargs):
        start_time = datetime.datetime.now()
        tmp = fn(*args, **kwargs)
        end_time = datetime.datetime.now()
        duration = end_time - start_time
        print(f"[{fn.__module__}.{fn.__qualname__}] Execution time: {duration.microseconds / 1000} ms")
        return tmp

    return wrapper



import sys
import os


def module_fn_definition(fn) -> str:
    return f"{fn.__module__}.{fn.__name__}"


def parameters_fn_definition(*args, **kwargs) -> str:
    return ', '.join(
        [parameter_definition(key, arg) for key, arg in enumerate(args)] +
        [parameter_definition(name, kwarg) for name, kwarg in kwargs.items()]
    )


def parameter_definition(name, value) -> str:
    return f"{name}: {repr(value)} -> {type(value)}"


def debug(fn):
    """
    MODULE NAME
        debug

    MODULE REFERENCE
        https://decoratorutilities.readthedocs.io/en/latest/rst_templates/Decorators/debug_decorator.html

    Decorate your own function with **@debug** decorator to print in console more Exception details

    :param fn: Decorated function or class method to debug
    :return:
    """
    from decoratorutilities import logger

    def wrapper(*args, **kwargs):
        base_string = f"{module_fn_definition(fn)}({parameters_fn_definition(*args, **kwargs)})"
        
        try:
            retr = fn(*args, **kwargs)
        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()

            fname = os.path.sep.join(os.path.split(exc_tb.tb_next.tb_frame.f_code.co_filename))
            if fname.startswith(sys.path[0]):
                fname = fname.replace(sys.path[0], "")

            logger.info(f"[{fname}:{exc_tb.tb_next.tb_lineno}]", base_string, "throws", parameter_definition("error", e))

            raise e
        else:
            logger.info(base_string, parameter_definition("return", retr))
            return retr

    return wrapper



from functools import wraps

memoized_value_handlers_map = {}
__all__ = ['memoized']


def get_value_matching(memoized_value_handlers_map, args_tuple):
    args, kwargs = args_tuple
    for value_handler in memoized_value_handlers_map:
        if args == value_handler["args"] and kwargs == value_handler["kwargs"]:
            return value_handler
    return None


def memoized(fn):
    """
    MODULE NAME
        memoized

    MODULE REFERENCE
        https://decoratorutilities.readthedocs.io/en/latest/rst_templates/Decorators/memoized_decorator.html

    Decorate your own function or class method with **@memoized** decorator to speed up it by storing the results and returning the cached result when the same inputs occur again

    :param fn: Instance of decorated function or class method
    :return: Return the execution of decorated function or class method
    """
    fn_name = f"{fn.__module__}.{fn.__qualname__}"

    if fn_name not in memoized_value_handlers_map:
        memoized_value_handlers_map[fn_name] = []

    @wraps(fn)
    def wrapper(*args, **kwargs):

        check = get_value_matching(memoized_value_handlers_map[fn_name], (args, kwargs))
        if not check:
            tmp = fn(*args, **kwargs)
            memoized_value_handlers_map[fn_name].append({"args": args, "kwargs": kwargs, "return": tmp})
            return tmp
        else:
            return check["return"]

    return wrapper



from ._checktype import checktype
from ._overload import overload
from ._mocking import mocking
from ._memoized import memoized
from ._timeit import timeit
from ._singleton import singleton
from ._debug import debug

import logging

logging.basicConfig(level=logging.DEBUG, format='%(asctime)s %(message)s', datefmt='%m/%d/%Y %I:%M:%S %p')
logger = logging.getLogger("decoratorutilities")


# FIXME: Da rivedere e correggere
#def is_class_method(arg):
#    Controllo se il parametro passato è una funzione o un metodo di una classe - args[0]
#    return "__module__" in dir(arg) and arg.__module__ is not None



import pytest
from decoratorutilities import overload

# modulo # scope # nome funzione


def test_with_mixing_positions_name_param():

    @overload
    def x(a: int, b: str, c: dict):
        return 1

    @overload
    def x(a: int):
        return 2

    assert x(c={}, b='stringa', a=0) == 1
    assert x(a=0, b='stringa', c={}) == 1
    assert x(0, 'stringa', {}) == 1

    assert x(1) == 2
    assert x(a=3) == 2


def test_with_missing_type():

    with pytest.raises(TypeError):
        @overload
        def x(a):
            pass


def test_partial_overload_types():

    @overload
    def a(a:int, b):
        return 1

    @overload
    def a(a:str, b):
        return 2

    assert a(1, None) == 1
    assert a('1', None) == 2


def test_calling_no_matching_signature():

    @overload
    def a(a: int):
        return 1

    with pytest.raises(ValueError):
        a('string')


def test_base_overload():

    @overload
    def a(a: int, b: str) -> int:
        return 0

    @overload
    def a(a: str) -> str:
        return "hello, world!"

    @overload
    def a() -> str:
        return "Senza params!"

    @overload
    def b(a: int) -> str:
        return "Funzione B!"

    @overload
    def b(a: str, b: int) -> int:
        return 5

    assert a(5, b="10") == 0
    assert a("1") == "hello, world!"
    assert a() == "Senza params!"
    assert b(0) == "Funzione B!"
    assert b("Ciao", 1) == 5


def test_calling_missing_signature_method():

    @overload
    def a(a: int):
        return 0

    @overload
    def a(a: str):
        return 1

    with pytest.raises(ValueError):
        assert a([1]) is None
        assert a([1, 2, 3, 4, 5]) == 5


def test_defining_same_signature_function():

    @overload
    def a(a: int):
        return 0

    with pytest.raises(NameError):
        @overload
        def a(a: int):
            return 0


def test_overload_with_class_methods():

    class X:
        @overload
        def x(self, x: int):
            return int

        @overload
        def x(self, x: str):
            return str

    assert X().x(1) == int
    assert X().x('1') == str



from functools import wraps

__all__ = ['checktype']


def checktype(fn):
    """
    MODULE NAME
        checktype

    MODULE REFERENCE
        https://decoratorutilities.readthedocs.io/en/latest/rst_templates/Decorators/check_type_decorator.html

    Decorate your own function or class method with **@checktype** decorator to check parameters type

    :param fn: Decorated function or class method to check parameters for
    :return:
    """

    @wraps(fn)
    def wrapper(*args, **kwargs):

        cls = None

        if len(args) > 0:
            if args[0].__class__.__qualname__ == ".".join(fn.__qualname__.split(".")[:-1]):
                cls, *args = args

            for value, (_key, _type) in zip(args, fn.__annotations__.items()):
                if not isinstance(value, _type):
                    raise TypeError(f"Got: \'{value}\' of {type(value)} instance, expected {_type} instance for \"{_key}\" parameter")

        if len(kwargs) > 0:
            for kwarg in kwargs:
                if not isinstance(kwargs[kwarg], fn.__annotations__[kwarg]):
                    raise TypeError(f"Got: \'{kwargs[kwarg]}\' of {type(kwargs[kwarg])} instance, expected {fn.__annotations__[kwarg]} instance for \"{kwarg}\" parameter")

        if cls:
            fn_return = fn(cls, *args, **kwargs)
        else:
            fn_return = fn(*args, **kwargs)

        if "return" in fn.__annotations__ and not isinstance(fn_return, fn.__annotations__["return"]):
            raise TypeError(
                f"Got: \'{type(fn_return)}\' return type, expected {fn.__annotations__['return']} return type for \"{fn.__name__}\" function")

        return fn_return
    return wrapper



import pytest

from decoratorutilities import mocking


def test_mocking():

    @mocking([
        ((1, 2, 3), {"a": 1}, 1),
        ((4, 5, 6), {"b": 2}, 2)
    ])
    def a():
        pass

    assert a(1, 2, 3, a=1) == 1
    assert a(4, 5, 6, b=2) == 2


def test_raise_missing_mocking():
    @mocking([
        ((7, 8, 9), {"c": 3}, 3)
    ])
    def a():
        pass

    @mocking([])
    def b():
        pass

    with pytest.raises(KeyError):
        a(1)

    with pytest.raises(KeyError):
        a(7, 8, 9, c=1)


def test_raise_missing_mocking_parameters():

    with pytest.raises(TypeError):
        @mocking([
            ((7, 8, 9), {"c": 3})
        ])
        def a():
            pass

    with pytest.raises(TypeError):
        @mocking([
            ({7, 8, 9}, {"c": 3})
        ])
        def a():
            pass


def test_mocking_wrong_argument_decorator():

    with pytest.raises(TypeError):
        @mocking({1, 2, 3, 4})
        def a():
            pass


def test_mocking_same_signature_different_return():

    with pytest.raises(KeyError):
        @mocking([
            ((1, 2, 3), {"a": 1}, 1),
            ((1, 2, 3), {"a": 1}, 2)
        ])
        def a():
            pass




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="decoratorutilities",
    version="1.0.2",
    author="Scalamandrè Simone, Giovanni Cardamone",
    author_email="simone.scalamandre@sourcesense.com, giovanni.cardamone@sourcesense.com",
    description="Python library to user type guard utilities to check parameters and return type, allow function overloading and function mocking at runtime",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Symone95/decoratorutilities",
    download_url="https://pypi.org/project/decoratorutilities/#files",
    packages=["decoratorutilities"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',

)



from functools import wraps
__all__ = ['singleton']


class Singleton(object):
    """
    Decorate your own classes with **@singleton()** decorator to ensure that only one instance of the singleton class ever exists.
    Never invoke **@singleton()** decorator without brackets otherwise it will cause problems
    Define your class method `__init__()` without parameter, pass them to the **@singleton()** decorator in the format "key" = "value" like kwargs.
    """

    def __init__(self, klass, *args, **kwargs):
        self.args = {}
        self.klass_name = klass.__name__
        if len(args) > 0:
            for index, arg in enumerate(args):
                self.__setitem__(str(index), arg)

        if len(kwargs) > 0:
            for kwarg in kwargs:
                setattr(self, kwarg, kwargs[kwarg])

        obj = klass()
        self.__dict__.update(obj.__dict__)  # Merge args and kwargs with Singleton object

        # Merge decorated class methods with Singleton object
        object_methods = [method_name for method_name in dir(obj)
                          if callable(getattr(obj, method_name))]
        for m in object_methods:
            if not m.startswith("__"):
                setattr(self, m, getattr(obj, m))

    def __call__(self, *args, **kwargs):
        raise TypeError(f'\"{self.klass_name}\" object is a singleton not instantiable')

    def __setitem__(self, key, value):
        self.args[key] = value

    def __getitem__(self, item):
        # If item is digit return item in "self.keys" dict with that index
        if isinstance(item, int):
            if item <= len(self.args):
                return self.args[str(item)]
            else:
                raise IndexError(f"Index: \"{item}\" not found in {self.klass_name} class")
        elif isinstance(item, str):
            if item in self.args:
                return self.args[item]
            else:
                raise KeyError(f"Key: \"{item}\" not found in {self.klass_name} class")

    def __delitem__(self, key):
        del self.args[key]

    def __str__(self):
        return self.klass_name

    def __repr__(self):
        return "%s(%r)" % (self.__class__, self.__dict__)

    def __enter__(self):
        self.fp = open(self.filename, "a+")
        return self.instance

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.fp.close()


def singleton(*args, **kwargs):
    """
    MODULE NAME
        singleton

    MODULE REFERENCE
        https://decoratorutilities.readthedocs.io/en/latest/rst_templates/Decorators/singleton_decorator.html

    Decorate your own classes with **@singleton()** decorator to ensure that only one instance of the singleton class ever exists.
    Never invoke **@singleton()** decorator without brackets otherwise it will cause problems
    Define your class method `__init__()` without parameter, pass them to the **@singleton()** decorator in the format "key" = "value" like kwargs.

    :param args: Class method *args
    :param kwargs: Class method **kwargs
    :return: Singleton instance of decorated class
    :raise: TypeError if klass is not a class method
    """

    @wraps(Singleton)
    def class_wrapper(klass):

        if not str(klass).startswith("<class "):
            raise TypeError("Singleton decorator applicable on class only")
        return Singleton(klass, *args, **kwargs)  # wrapper()
    return class_wrapper



from functools import wraps
from ._checktype import checktype


__all__ = ['mocking']

value_handler = []


@checktype
def save_mock(args: tuple, kwargs: dict, retr):
    mock = (args, kwargs, retr)
    if mock not in value_handler:
        value_handler.append(mock)
    else:
        raise KeyError(f"Parameters {mock} signature already exists")


def is_mock_matching(args_tuple, mocking_value_handler):
    args, kwargs = args_tuple

    for tuple_handler in mocking_value_handler:
        if args in tuple_handler and kwargs in tuple_handler:
            return tuple_handler
    return None


@checktype
def mocking(mocks: list):
    """
    MODULE NAME
        mocking

    MODULE REFERENCE
        https://decoratorutilities.readthedocs.io/en/latest/rst_templates/Decorators/mocking_decorator.html

    Decorate your own function with **@mocking** decorator to mock that function adding args in a tuple, kwargs in a dict and return value

    :param mocks: List contains mocked *args, **kwargs and return
    :return:
    """

    for mock in mocks:
        save_mock(*mock)

    def fn_wrapper(fn):

        @wraps(fn)
        def arg_wrapper(*args, **kwargs):
            args_tuple = is_mock_matching((args, kwargs), value_handler)
            if args_tuple:
                return args_tuple[2]
            else:
                raise KeyError("Unmatched parameters: ", (args, kwargs), " -> choose between: ", value_handler)

        return arg_wrapper

    return fn_wrapper



import pytest
from decoratorutilities import checktype


def test_no_annotation():

    @checktype
    def a(a):
        return a

    assert a(1) == 1


def test_base_param():

    @checktype
    def a(a: int, b: int):
        return 1

    assert a(5, 6) == 1


def test_base_params():
    @checktype
    def a(a: int, b: int):
        return 1

    with pytest.raises(TypeError):
        a("invalid", "Invalid")

    with pytest.raises(TypeError):
        a("invalid", b="Invalid")

    with pytest.raises(TypeError):
        a(a="invalid", b="Invalid")


def test_pos_named_param():

    @checktype
    def a(a: str, b: int, c: list):
        return 1

    with pytest.raises(TypeError):
        a("Valid", 1, None)

    with pytest.raises(TypeError):
        a(a="Valid", b=1, c=None)

    assert a("Valid", 1, []) == 1
    assert a(a="Valid", b=1, c=[]) == 1
    assert a(c=[], b=1, a="Valid") == 1


def test_base_return():

    @checktype
    def a() -> int:
        return "Invalid"

    with pytest.raises(TypeError):
        a()


def test_valid_return():

    @checktype
    def a() -> int:
        return 0

    assert a() == 0

glob = 0


def test_unique_call():

    @checktype
    def a() -> int:
        global glob
        glob += 1
        return glob

    assert a() == 1
    assert a() == 2


def test_class_arguments():

    class X(object):

        @checktype
        def x(self, value: int):
            return value

    assert X().x(1) == 1

    with pytest.raises(TypeError):
        X().x('1')
